import { useState, memo, useMemo, useEffect, Component } from 'react';
import { observe, unobserve, isObservable, raw, observable } from '@nx-js/observer-util';

let batchedUpdates;

try {
  const ReactDOM = require('react-dom');

  batchedUpdates = ReactDOM.unstable_batchedUpdates;
} catch (err) {
  try {
    const ReactNative = require('react-native');

    batchedUpdates = ReactNative.unstable_batchedUpdates;
  } catch (err) {
    batchedUpdates = fn => fn();
  }
} // try to find the global object
// it is window in the DOM and global in NodeJS and React Native


const isDOM = typeof window !== 'undefined';
const isNative = typeof global !== 'undefined';
const globalObj = isDOM ? window : isNative ? global : undefined;
const hasHooks = typeof useState === 'function';

let isInsideFunctionComponent = false;
const COMPONENT = Symbol('owner component');
const DUMMY_STATE = {};
function view(Comp) {
  const isStatelessComp = !(Comp.prototype && Comp.prototype.isReactComponent);
  let ReactiveComp;

  if (isStatelessComp && hasHooks) {
    // use a hook based reactive wrapper when we can
    ReactiveComp = memo(props => {
      // create a memoized reactive wrapper of the original component (render)
      // at the very first run of the component function
      const render = useMemo(() => {
        const [, setState] = useState();
        return observe(Comp, {
          scheduler: () => setState(DUMMY_STATE),
          lazy: true
        });
      }, []); // cleanup the reactive connections after the very last render of the component

      useEffect(() => {
        return () => unobserve(render);
      }, []); // the isInsideFunctionComponent flag is used to toggle `store` behavior
      // based on where it was called from

      isInsideFunctionComponent = true;

      try {
        // run the reactive render instead of the original one
        return render(props);
      } finally {
        isInsideFunctionComponent = false;
      }
    });
  } else {
    const BaseComp = isStatelessComp ? Component : Comp; // a HOC which overwrites render, shouldComponentUpdate and componentWillUnmount
    // it decides when to run the new reactive methods and when to proxy to the original methods

    class ReactiveClassComp extends BaseComp {
      constructor(props, context) {
        super(props, context);
        this.state = this.state || {};
        this.state[COMPONENT] = this; // create a reactive render for the component

        this.render = observe(this.render, {
          scheduler: () => this.setState(DUMMY_STATE),
          lazy: true
        });
      }

      render() {
        return isStatelessComp ? Comp(this.props, this.context) : super.render();
      } // react should trigger updates on prop changes, while easyState handles store changes


      shouldComponentUpdate(nextProps, nextState) {
        const {
          props,
          state
        } = this; // respect the case when the user defines a shouldComponentUpdate

        if (super.shouldComponentUpdate) {
          return super.shouldComponentUpdate(nextProps, nextState);
        } // return true if it is a reactive render or state changes


        if (state !== nextState) {
          return true;
        } // the component should update if any of its props shallowly changed value


        const keys = Object.keys(props);
        const nextKeys = Object.keys(nextProps);
        return nextKeys.length !== keys.length || nextKeys.some(key => props[key] !== nextProps[key]);
      } // add a custom deriveStoresFromProps lifecyle method


      static getDerivedStateFromProps(props, state) {
        if (super.deriveStoresFromProps) {
          // inject all local stores and let the user mutate them directly
          const stores = mapStateToStores(state);
          super.deriveStoresFromProps(props, ...stores);
        } // respect user defined getDerivedStateFromProps


        if (super.getDerivedStateFromProps) {
          return super.getDerivedStateFromProps(props, state);
        }

        return null;
      }

      componentWillUnmount() {
        // call user defined componentWillUnmount
        if (super.componentWillUnmount) {
          super.componentWillUnmount();
        } // clean up memory used by Easy State


        unobserve(this.render);
      }

    }

    ReactiveComp = ReactiveClassComp;
  }

  ReactiveComp.displayName = Comp.displayName || Comp.name; // static props are inherited by class components,
  // but have to be copied for function components

  if (isStatelessComp) {
    for (let key of Object.keys(Comp)) {
      ReactiveComp[key] = Comp[key];
    }
  }

  return ReactiveComp;
}

function mapStateToStores(state) {
  // find store properties and map them to their none observable raw value
  // to do not trigger none static this.setState calls
  // from the static getDerivedStateFromProps lifecycle method
  const component = state[COMPONENT];
  return Object.keys(component).map(key => component[key]).filter(isObservable).map(raw);
}

function store(obj) {
  // do not create new versions of the store on every render
  // if it is a local store in a function component
  // create a memoized store at the first call instead
  if (hasHooks && isInsideFunctionComponent) {
    return useMemo(() => observable(obj), []);
  }

  return observable(obj);
}

/* eslint camelcase: 0 */
// until the function is finished running

function batch(fn, ctx, args) {
  let result;
  batchedUpdates(() => result = fn.apply(ctx, args));
  return result;
} // this creates and returns a batched version of the passed function
// the cache is necessary to always map the same thing to the same function
// which makes sure that addEventListener/removeEventListener pairs don't break

const cache = new Map();

function batchFn(fn) {
  if (typeof fn !== 'function') {
    return fn;
  }

  let batched = cache.get(fn);

  if (!batched) {
    batched = function (...args) {
      return batch(fn, this, args);
    };

    cache.set(fn, batched);
  }

  return batched;
} // batched window.addEventListener(cb) like callbacks


function batchCallbacks(functionWithCallbacks) {
  return function batchedCallbacks(...args) {
    return functionWithCallbacks.apply(this, args.map(batchFn));
  };
} // batches obj.onevent = fn like calls


function batchMethod(obj, method) {
  const descriptor = Object.getOwnPropertyDescriptor(obj, method);

  if (descriptor) {
    const newDescriptor = Object.assign({}, descriptor, {
      set(value) {
        return descriptor.set.call(this, batchFn(value));
      }

    });
    Object.defineProperty(obj, method, newDescriptor);
  }
} // do a sync batching for the most common task sources
// this should be removed when React's own batching is improved in the future


if (globalObj) {
  // batch timer functions
  globalObj.setTimeout = batchCallbacks(globalObj.setTimeout);
  globalObj.setInterval = batchCallbacks(globalObj.setInterval);

  if (globalObj.requestAnimationFrame) {
    globalObj.requestAnimationFrame = batchCallbacks(globalObj.requestAnimationFrame);
  }

  if (globalObj.requestIdleCallback) {
    globalObj.requestIdleCallback = batchCallbacks(globalObj.requestIdleCallback);
  } // eslint-disable-next-line


  Promise.prototype.then = batchCallbacks(Promise.prototype.then); // eslint-disable-next-line

  Promise.prototype.catch = batchCallbacks(Promise.prototype.catch); // batch addEventListener calls

  if (globalObj.EventTarget) {
    EventTarget.prototype.addEventListener = batchCallbacks(EventTarget.prototype.addEventListener);
    EventTarget.prototype.removeEventListener = batchCallbacks(EventTarget.prototype.removeEventListener);
  } // this batches websocket event handlers


  if (globalObj.WebSocket) {
    const websocketHandlers = ['onopen', 'onmessage', 'onerror', 'onclose'];
    websocketHandlers.forEach(method => batchMethod(globalObj.WebSocket.prototype, method));
  } // HTTP event handlers are usually wrapped by Promises, which is covered above

}

export { view, store, batch };
//# sourceMappingURL=es.es6.js.map
